"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.trimAndRedact = exports.flushFaults = exports.faults = exports.FAULT_EVENT_TYPE = void 0;

var _highland = _interopRequireDefault(require("highland"));

var uuid = _interopRequireWildcard(require("uuid"));

var _cloneDeepWith = _interopRequireDefault(require("lodash/cloneDeepWith"));

var _utils = require("../utils");

const FAULT_EVENT_TYPE = 'fault'; // collect faults until the end so that an unhandled error
// does not cause handler faults to be repeatedly published

exports.FAULT_EVENT_TYPE = FAULT_EVENT_TYPE;
const theFaults = [];

const faults = (err, push) => {
  if (err.uow) {
    // handled exceptions are adorned with the uow in error
    // push a fault event onto the stack for publishing by publishFaultsPipeline
    theFaults.push({
      id: uuid.v1(),
      partitionKey: uuid.v4(),
      type: FAULT_EVENT_TYPE,
      timestamp: (0, _utils.now)(),
      tags: {
        functionname: process.env.AWS_LAMBDA_FUNCTION_NAME || 'undefined',
        pipeline: err.uow.pipeline || 'undefined'
      },
      err: {
        name: err.name,
        message: err.message,
        stack: err.stack
      },
      uow: trimAndRedact(err.uow)
    });
    logErr(err);
  } else {
    // rethrow unhandled/unexpected exceptions to stop processing
    push(err);
  }
};

exports.faults = faults;

const flushFaults = opt => s => {
  // use at the every end with through() to redirect to faults streams
  const faultStream = () => {
    const s2 = (0, _highland.default)((push, next) => {
      const f = theFaults.shift();

      if (f) {
        push(null, f);
        next();
      } else {
        push(null, _highland.default.nil);
      }
    });
    return s2 // batch and publish fault events
    .map(fault => ({
      event: fault
    })) // map to uow format
    .through(opt.publish({ ...opt,
      ...opt.faultOpt,
      // override options specific for faults
      handleErrors: false,
      // don't publish faults for faults
      batchSize: Number(process.env.FAULTS_BATCH_SIZE) || Number(process.env.BATCH_SIZE) || 4,
      parallel: Number(process.env.FAULTS_PARALLEL) || Number(process.env.PARALLEL) || 4
    }));
  };

  return s.consume((err, x, push, next) => {
    /* istanbul ignore if */
    if (err) {
      push(err);
      next();
    } else if (x === _highland.default.nil) {
      // this is the purpose of this consume step
      // publish all acuumulated faults at the very end
      next(faultStream());
    } else {
      push(null, x);
      next();
    }
  });
};

exports.flushFaults = flushFaults;

const logErr = err => {
  /* istanbul ignore if */
  if (process.env.AWS_LAMBDA_LOG_GROUP_NAME) {
    if (err instanceof Error) {
      console.error(JSON.stringify({
        errorMessage: err.message,
        errorType: err.name,
        pipeline: err.uow.pipeline || 'undefined',
        handled: err.uow !== undefined,
        stackTrace: err.stack
      }));
    } else {
      console.error(err);
    }
  }
};

const trimAndRedact = _uow => {
  const fields = (_uow.batch || [_uow]).reduce((a, c) => {
    var _c$event, _c$undecryptedEvent;

    const eem = ((_c$event = c.event) === null || _c$event === void 0 ? void 0 : _c$event.eem) || ((_c$undecryptedEvent = c.undecryptedEvent) === null || _c$undecryptedEvent === void 0 ? void 0 : _c$undecryptedEvent.eem);
    const f = (eem === null || eem === void 0 ? void 0 : eem.fields) || [];
    return [...a, ...f];
  }, []);
  const cache = [];

  const cloneCustomizer = (value, key) => {
    if (fields.includes(key)) {
      return '[REDACTED]';
    } else {
      if (Buffer.isBuffer(value)) {
        return `[BUFFER: ${value.length}]`;
      }

      if (typeof value === 'object' && value !== null) {
        if (cache.includes(value)) {
          return '[CIRCULAR]';
        } else {
          cache.push(value);
        }
      }
    }

    return undefined;
  };

  const tr = uow => {
    const {
      pipeline,
      record,
      event,
      decryptResponse,
      undecryptedEvent,
      encryptResponse,
      ...rest
    } = uow;
    return {
      pipeline,
      record,
      // DO NOT redact so we can resubmit
      ...(0, _cloneDeepWith.default)({
        event: undecryptedEvent || event,
        ...rest
      }, cloneCustomizer)
    };
  };

  if (!_uow.batch) {
    return tr(_uow);
  } else {
    const {
      batch,
      ...rest2
    } = _uow;
    return {
      batch: _uow.batch.map(u => tr(u)),
      ...(0, _cloneDeepWith.default)({ ...rest2
      }, cloneCustomizer)
    };
  }
};

exports.trimAndRedact = trimAndRedact;