"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.collect = void 0;

var _get = _interopRequireDefault(require("lodash/get"));

var _isFunction = _interopRequireDefault(require("lodash/isFunction"));

var _omit = _interopRequireDefault(require("lodash/omit"));

var _utils = require("../utils");

var _filters = require("../filters");

var _dynamodb = require("../utils/dynamodb");

/**
 * collects events in a micro event store
 * used in listener functions
 * provides idempotency
 * the partitionKey is used as the default correlation key
 * uses the DynamoDB single table pattern
 *
 * interface Rule {
 *   id: string
 *   flavor: collect,
 *   eventType: string | string[] | Function,
 *   filters?: Function[],
 *   correlationKey?: string | Function, // default uow.event.partitionKey
 *   ttl?: number, // default 33 days
 *   expire?: boolean | string
 *   parallel?: number;
 *   toPutRequest?: Function,
 * }
 */
const collect = rule => s => s // eslint-disable-line import/prefer-default-export
.filter(_filters.outSkip).filter(onEventType(rule)).tap(_utils.printStartPipeline).filter(onContent(rule)).map(correlationKey(rule)).map(toPutRequest(rule)).through((0, _dynamodb.putDynamoDB)(rule)).tap(_utils.printEndPipeline);

exports.collect = collect;

const onEventType = rule => (0, _utils.faulty)(uow => (0, _filters.filterOnEventType)(rule, uow));

const onContent = rule => (0, _utils.faulty)(uow => (0, _filters.filterOnContent)(rule, uow));

const correlationKey = rule => (0, _utils.faulty)(uow => {
  let key;

  if (!rule.correlationKey) {
    key = uow.event.partitionKey;
  } else if ((0, _isFunction.default)(rule.correlationKey)) {
    key = rule.correlationKey(uow);
  } else {
    key = (0, _get.default)(uow.event, rule.correlationKey);
  }

  return { ...uow,
    key
  };
});

const toPutRequest = rule => (0, _utils.faulty)(uow => {
  var _uow$record$kinesis, _uow$record$attribute;

  return { ...uow,
    putRequest: // variable expected by `put` util
    rule.toPutRequest ?
    /* istanbul ignore next */
    rule.toPutRequest(uow, rule) : {
      Item: {
        pk: uow.event.id,
        sk: 'EVENT',
        discriminator: 'EVENT',
        timestamp: uow.event.timestamp,
        awsregion: process.env.AWS_REGION,
        sequenceNumber: ((_uow$record$kinesis = uow.record.kinesis) === null || _uow$record$kinesis === void 0 ? void 0 : _uow$record$kinesis.sequenceNumber) || (
        /* istanbul ignore next */
        (_uow$record$attribute = uow.record.attributes) === null || _uow$record$attribute === void 0 ? void 0 : _uow$record$attribute.SequenceNumber),
        ttl: (0, _utils.ttlRule)(rule, uow),
        expire: rule.expire,
        data: uow.key,
        event: rule.includeRaw ?
        /* istanbul ignore next */
        uow.event : (0, _omit.default)(uow.event, ['raw'])
      }
    }
  };
});